import logging
log = logging.getLogger('onegov.user')  # noqa
log.addHandler(logging.NullHandler())  # noqa

from translationstring import TranslationStringFactory
_ = TranslationStringFactory('onegov.user')  # noqa

from onegov.user.auth import Auth
from onegov.user.collections import UserCollection
from onegov.user.collections import UserGroupCollection
from onegov.user.integration import UserApp
from onegov.user.models import User
from onegov.user.models import UserGroup

__all__ = [
    'Auth',
    'User',
    'UserApp',
    'UserGroup',
    'UserCollection',
    'UserGroupCollection'
]



from onegov.user.models import UserGroup
from onegov.core.collection import GenericCollection


class UserGroupCollection(GenericCollection):
    """ Manages a list of user groups.

    Use it like this::

        from onegov.user import UserGroupCollection
        groups = UserGroupCollection(session)

    """

    def __init__(self, session, type='*'):
        self.session = session
        self.type = type

    @property
    def model_class(self):
        return UserGroup.get_polymorphic_class(self.type, default=UserGroup)

    def query(self):
        query = super(UserGroupCollection, self).query()
        return query.order_by(self.model_class.name)



from freezegun import freeze_time
from onegov.user import User
from onegov.user import UserGroup
from unittest.mock import call
from unittest.mock import patch


class DummyBrowserSession():
    def __init__(self, token):
        self._token = token

    def flush(self):
        pass


class DummyRequest():
    def __init__(self, token):
        self.client_addr = '127.0.0.1'
        self.user_agent = 'Mozilla/5.0'
        self.browser_session = DummyBrowserSession(token)
        self.app = None


def test_user_initials():
    user = User(username='admin')
    assert user.initials == 'A'

    user = User(username='nathan.drake@example.org')
    assert user.initials == 'ND'

    user = User(username='a.b.c.d.e.f')
    assert user.initials == 'AB'

    user = User(username='victor', realname='Victor Sullivan')
    assert user.initials == 'VS'

    user = User(username='burns', realname='Charles Montgomery Burns')
    assert user.initials == 'CB'


def test_user_title(session):
    user = User(username='admin')
    assert user.title == 'admin'

    user = User(username='nathan', realname="Nathan Drake")
    assert user.title == "Nathan Drake"


def test_user_group(session):
    session.add(User(username='nathan', password='pwd', role='editor'))
    session.add(UserGroup(name='group'))
    session.flush()
    user = session.query(User).one()
    group = session.query(UserGroup).one()

    assert user.group is None
    assert group.users.all() == []

    user.group = group
    assert group.users.one() == user
    assert user.group == group
    assert user.group_id == group.id


def test_polymorphism_user(session):

    class MyUser(User):
        __mapper_args__ = {'polymorphic_identity': 'my'}

    class MyOtherUser(User):
        __mapper_args__ = {'polymorphic_identity': 'other'}

    session.add(User(username='default', password='pwd', role='editor'))
    session.add(MyUser(username='my', password='pwd', role='editor'))
    session.add(MyOtherUser(username='other', password='pwd', role='editor'))
    session.flush()

    assert session.query(User).count() == 3
    assert session.query(MyUser).one().username == 'my'
    assert session.query(MyOtherUser).one().username == 'other'


def test_polymorphism_group(session):

    class MyGroup(UserGroup):
        __mapper_args__ = {'polymorphic_identity': 'my'}

    class MyOtherGroup(UserGroup):
        __mapper_args__ = {'polymorphic_identity': 'other'}

    session.add(UserGroup(name='original'))
    session.add(MyGroup(name='my'))
    session.add(MyOtherGroup(name='other'))
    session.flush()

    assert session.query(UserGroup).count() == 3
    assert session.query(MyGroup).one().name == 'my'
    assert session.query(MyOtherGroup).one().name == 'other'


def test_user_save_session():
    user = User()
    assert not user.sessions

    with patch('onegov.user.models.user.remembered') as remembered:
        remembered.return_value = True

        with freeze_time('2016-09-01 12:00'):
            user.save_current_session(DummyRequest('xxx'))

        with freeze_time('2016-10-01 12:00'):
            user.save_current_session(DummyRequest('yyy'))

        assert user.sessions == {
            'xxx': {
                'address': '127.0.0.1',
                'timestamp': '2016-09-01T12:00:00',
                'agent': 'Mozilla/5.0'
            },
            'yyy': {
                'address': '127.0.0.1',
                'timestamp': '2016-10-01T12:00:00',
                'agent': 'Mozilla/5.0'
            }
        }


def test_user_remove_session():
    user = User()
    assert not user.sessions

    with patch('onegov.user.models.user.remembered') as remembered:
        remembered.return_value = True

        user.save_current_session(DummyRequest('xxx'))
        user.save_current_session(DummyRequest('yyy'))
        assert 'xxx' in user.sessions
        assert 'yyy' in user.sessions

        user.remove_current_session(DummyRequest('xxx'))
        assert 'xxx' not in user.sessions
        assert 'yyy' in user.sessions


def test_user_logout_all_sessions():
    user = User()
    assert not user.sessions

    with patch('onegov.user.models.user.remembered') as remembered:
        with patch('onegov.user.models.user.forget') as forget:
            remembered.return_value = True

            user.save_current_session(DummyRequest('xxx'))
            user.save_current_session(DummyRequest('yyy'))
            assert 'xxx' in user.sessions
            assert 'yyy' in user.sessions

            user.logout_all_sessions(DummyRequest('zzz'))
            assert call(None, 'xxx') in forget.mock_calls
            assert call(None, 'yyy') in forget.mock_calls


def test_user_cleanup_sessions():
    user = User()
    assert not user.sessions

    with patch('onegov.user.models.user.remembered') as remembered:
        with patch('onegov.user.models.user.forget'):
            # ... implicit
            remembered.return_value = True
            user.save_current_session(DummyRequest('xxx'))
            assert 'xxx' in user.sessions

            remembered.return_value = False
            user.cleanup_sessions(DummyRequest('zzz'))
            assert 'xxx' not in user.sessions

            # ... while adding
            remembered.return_value = True
            user.save_current_session(DummyRequest('xxx'))
            assert 'xxx' in user.sessions

            remembered.return_value = False
            user.save_current_session(DummyRequest('yyy'))
            assert 'xxx' not in user.sessions
            assert 'yyy' not in user.sessions

            # ... while removing
            remembered.return_value = True
            user.save_current_session(DummyRequest('xxx'))
            user.save_current_session(DummyRequest('yyy'))
            assert 'xxx' in user.sessions
            assert 'yyy' in user.sessions

            remembered.return_value = False
            user.remove_current_session(DummyRequest('xxx'))
            assert 'xxx' not in user.sessions
            assert 'yyy' not in user.sessions

            # ... while logging out
            remembered.return_value = True
            user.save_current_session(DummyRequest('xxx'))
            user.save_current_session(DummyRequest('yyy'))
            assert 'xxx' in user.sessions
            assert 'yyy' in user.sessions

            remembered.return_value = False
            user.logout_all_sessions(DummyRequest('zzz'))
            assert 'xxx' not in user.sessions
            assert 'yyy' not in user.sessions



import morepath

from datetime import datetime
from itsdangerous import URLSafeSerializer, BadSignature

from onegov.core.utils import relative_url
from onegov.user import log
from onegov.user.collections import UserCollection
from onegov.user.errors import ExpiredSignupLinkError
from onegov.user.auth.second_factor import SECOND_FACTORS


class Auth(object):
    """ Defines a model for authentication methods like login/logout.
    Applications should use this model to implement authentication views.

    """

    identity_class = morepath.Identity

    def __init__(self, session, application_id, to='/', skip=False,
                 signup_token=None, signup_token_secret=None, **kwargs):
        assert application_id  # may not be empty!

        self.session = session
        self.application_id = application_id

        self.signup_token = signup_token
        self.signup_token_secret = signup_token_secret

        # never redirect to an external page, this might potentially be used
        # to trick the user into thinking he's on our page after entering his
        # password and being redirected to a phishing site.
        self.to = relative_url(to)
        self.skip = skip

        # initialize nth factors
        self.factors = {}

        for type, cls in SECOND_FACTORS.items():
            obj = cls(**kwargs)

            if obj.is_configured():
                self.factors[type] = obj

    @classmethod
    def from_app(cls, app, to='/', skip=False, signup_token=None):
        kwargs = {}

        for factor in SECOND_FACTORS.values():
            kwargs.update(factor.args_from_app(app))

        return cls(
            session=app.session(),
            application_id=app.application_id,
            to=to,
            skip=skip,
            signup_token=signup_token,
            signup_token_secret=getattr(app, 'identity_secret', None),
            **kwargs,
        )

    @classmethod
    def from_request(cls, request, to='/', skip=False, signup_token=None):
        return cls.from_app(request.app, to, skip, signup_token)

    @classmethod
    def from_request_path(cls, request, skip=False, signup_token=None):
        return cls.from_request(
            request, request.transform(request.path), skip, signup_token)

    @property
    def users(self):
        return UserCollection(self.session)

    def redirect(self, request):
        return morepath.redirect(request.transform(self.to))

    def skippable(self, request):
        """ Returns true if the login for the current `to` target is optional
        (i.e. it is not required to access the page).

        This should only be used on protected pages as public pages would
        always be skipppable. Therefore it has to be enabled manually by
        specifying `skip=True` on the :class:`Auth` class.

        """

        if not self.skip:
            return False

        # this is the default parameter, we won't skip to it in any case
        if self.to == '/':
            return False

        try:
            return request.has_access_to_url(self.to)
        except KeyError:
            return False

    def is_valid_second_factor(self, user, second_factor_value):
        """ Returns true if the second factor of the given user is valid. """

        if not user.second_factor:
            return True

        if not second_factor_value:
            return False

        if user.second_factor['type'] in self.factors:
            return self.factors[user.second_factor['type']].is_valid(
                user_specific_config=user.second_factor['data'],
                factor=second_factor_value
            )
        else:
            raise NotImplementedError

    def authenticate(self, username, password,
                     client='unknown', second_factor=None):
        """ Takes the given username and password and matches them against the
        users collection. This does not login the user, use :meth:`login_to` to
        accomplish that.

        :param username:
            The username to authenticate.

        :param password:
            The password of the user (clear-text).

        :param client:
            The client address of the user (i.e. his IP address).

        :param second_factor:
            The value of the second factor or None.

        :return: The matched user, if successful, or None.

        """

        user = self.users.by_username_and_password(username, password)

        def fail():
            log.info(f"Failed login by {client} ({username})")
            return None

        if user is None:
            return fail()

        if not user.active:
            return fail()

        if user.authentication_provider:
            if user.authentication_provider.get('required', False):
                return fail()

        if not self.is_valid_second_factor(user, second_factor):
            return fail()

        log.info(f"Successful login by {client} ({username})")
        return user

    def as_identity(self, user):
        """ Returns the morepath identity of the given user. """

        return self.identity_class(
            userid=user.username,
            groupid=user.group_id.hex if user.group_id else '',
            role=user.role,
            application_id=self.application_id
        )

    def by_identity(self, identity):
        """ Returns the user record of the given identity. """

        return self.users.by_username(identity.userid)

    def login_to(self, username, password, request, second_factor=None):
        """ Takes a user login request and remembers the user if the
        authentication completes successfully.

        :param username:
            The username to log in.

        :param password:
            The password to log in (cleartext).

        :param request:
            The request of the user.

        :param second_factor:
            The second factor, if any.

        :return: A redirect response to ``self.to`` with the identity
        remembered as a cookie. If not successful, None is returned.

        """

        user = self.authenticate(
            username=username,
            password=password,
            client=request.client_addr,
            second_factor=second_factor)

        if user is None:
            return None

        return self.complete_login(user, request)

    def complete_login(self, user, request):
        """ Takes a user record, remembers its session and returns a proper
        redirect response to complete the login.

        This method is mostly useful inside onegov.user. You probably want
        to use :meth:`complete_login` outside of that.

        """
        assert user is not None

        identity = self.as_identity(user)
        response = self.redirect(request)

        request.app.remember_identity(response, request, identity)
        user.save_current_session(request)

        return response

    def logout_to(self, request):
        """ Logs the current user out and redirects to ``self.to``.

        :return: A response redirecting to ``self.to`` with the identity
        forgotten.

        """

        user = self.by_identity(request.identity)
        user and user.remove_current_session(request)

        response = self.redirect(request)
        request.app.forget_identity(response, request)

        return response

    def new_signup_token(self, role, max_age=24 * 60 * 60, max_uses=1):
        """ Returns a signup token which can be used for users to register
        themselves, directly gaining the given role.

        Signup tokens are recorded on the user to make sure that only the
        requested amount of uses is allowed.

        """
        return self.signup_token_serializer.dumps({
            'role': role,
            'max_uses': max_uses,
            'expires': int(datetime.utcnow().timestamp()) + max_age
        })

    @property
    def signup_token_serializer(self):
        assert self.signup_token_secret
        return URLSafeSerializer(self.signup_token_secret, salt='signup')

    def decode_signup_token(self, token):
        try:
            return self.signup_token_serializer.loads(token)
        except BadSignature:
            return None

    @property
    def permitted_role_for_registration(self):
        """ Returns the permitted role for the current signup token. """

        if not self.signup_token:
            return 'member'

        assert self.signup_token_secret
        params = self.decode_signup_token(self.signup_token)

        if not params:
            return None

        if params['expires'] < int(datetime.utcnow().timestamp()):
            return None

        signups = UserCollection(self.session)\
            .by_signup_token(self.signup_token).count()

        if signups >= params['max_uses']:
            return None

        return params['role']

    def register(self, form, request):
        """ Registers the user using the information on the registration form.

        Takes the signup token into account to provide the user with the
        proper role.

        See :meth:`onegov.user.collections.UserCollection.register_user` for
        more information.

        """

        role = self.permitted_role_for_registration

        if role is None:
            raise ExpiredSignupLinkError()

        return UserCollection(self.session).register(
            username=form.username.data,
            password=form.password.data,
            request=request,
            role=role,
            signup_token=self.signup_token
        )



from furl import furl


def password_reset_url(user, request, url):
    """ Appends the token needed by PasswordResetForm for a password reset.

    :user:
        The user (model).

    :request:
        The request.

    :url:
        The URL which points to the password reset view (which is using the
        PasswordResetForm).

    :return: An URL containg the password reset token.

    """

    token = request.new_url_safe_token({
        'username': user.username,
        'modified': user.modified.isoformat() if user.modified else ''
    })

    return furl(url).add({'token': token}).url



from onegov.user.collections.group import UserGroupCollection
from onegov.user.collections.user import MIN_PASSWORD_LENGTH
from onegov.user.collections.user import UserCollection


__all__ = [
    'MIN_PASSWORD_LENGTH',
    'UserCollection',
    'UserGroupCollection'
]



from onegov.form import Form
from onegov.user import _
from wtforms import StringField
from wtforms.validators import InputRequired


class UserGroupForm(Form):

    """ A generic user group form for onegov.user """

    name = StringField(
        label=_("Name"),
        validators=[
            InputRequired()
        ]
    )

    def update_model(self, model):
        model.name = self.name.data

    def apply_model(self, model):
        self.name.data = model.name



import base64
import kerberos
import morepath
import pytest
import transaction

from onegov.core import Framework
from onegov.core.utils import scan_morepath_modules
from onegov.user import Auth, UserApp, UserCollection
from onegov.user.auth.provider import provider_by_name
from tempfile import NamedTemporaryFile
from unittest.mock import patch, DEFAULT
from webtest import TestApp as Client


@pytest.fixture(scope='session')
def keytab():
    """ BASE 64 encoded keytab file for Kerberos integration tests

    Principal: HTTP/ogc.example.org@EXAMPLE.ORG
    Password: test

    To create, start ktutil (the latest release, macOS's one is too old):

        ktutil
        addent -password -p HTTP/ogc.example.org@EXAMPLE.ORG -k 1 -e aes256-cts
        wkt service.keytab
        exit
        cat service.keytab | base64

    """
    KEYTAB = (
        "BQIAAABXAAIAC0VYQU1QTEUuT1JHAARIVFRQAA9vZ2MuZXhhbXBsZS5vcmcAAAABXSxM"
        "KQEAEgAgKddJPBCQCDAtxV1NNksmnHT9xkbQLuO5rqFo+a6NEJMAAAAB"
    )

    with NamedTemporaryFile() as f:
        f.write(base64.b64decode(KEYTAB))
        f.flush()

        yield f.name


@pytest.fixture(scope='function')
def app(request, postgres_dsn, temporary_path, redis_url, keytab):

    class App(Framework, UserApp):
        pass

    scan_morepath_modules(App)
    morepath.commit(App)

    app = App()
    app.configure_application(
        dsn=postgres_dsn,
        redis_url=redis_url,
        authentication_providers={
            'kerberos': {
                'keytab': keytab,
                'hostname': 'ogc.example.org',
                'service': 'HTTP'
            }
        }
    )

    app.namespace = 'apps'
    app.set_application_id('apps/my-app')

    return app


@pytest.fixture(scope='function')
def client(app):
    yield Client(app)


def test_kerberos_auth(client, app):
    provider = provider_by_name(app.providers, 'kerberos')
    assert provider is not None

    def auth(headers=None):
        return client.get(
            url='/auth/provider/kerberos',
            expect_errors=True,
            headers=headers
        )

    # initially, the answer is 401 Unauthorized, to which Kerberos clients
    # are meant to answer with a ticket/token
    r = auth()
    assert r.status_code == 401
    assert r.headers['WWW-Authenticate'] == 'Negotiate'

    # for our tests we fake must mock the answers somewhat, setting up
    # a Kerberos environment is no easy task
    methods = {
        'authGSSServerInit': DEFAULT,
        'authGSSServerStep': DEFAULT,
        'authGSSServerResponse': DEFAULT,
        'authGSSServerUserName': DEFAULT,
    }
    with patch.multiple(kerberos, **methods) as methods:
        assert methods

        # simulate a failing init step
        methods['authGSSServerInit'].return_value = -1, None
        r = auth({'Authorization': 'Negotiate foobar'})

        assert r.status_code == 401
        assert r.headers['WWW-Authenticate'] == 'Negotiate'

        # simulate a failing challenge step
        methods['authGSSServerInit'].return_value = 1, None
        methods['authGSSServerStep'].return_value = -1
        r = auth({'Authorization': 'Negotiate foobar'})

        assert r.status_code == 401
        assert r.headers['WWW-Authenticate'] == 'Negotiate'

        # simulate a successful exchange, with no associated user
        methods['authGSSServerInit'].return_value = 1, None
        methods['authGSSServerStep'].return_value = 1
        methods['authGSSServerResponse'].return_value = 'barfoo'
        methods['authGSSServerUserName'].return_value = 'foo@EXAMPLE.ORG'
        r = auth({'Authorization': 'Negotiate foobar'})

        assert r.status_code == 403
        assert 'WWW-Authenticate' not in r.headers

        # associate a user, but keep it inactive
        user = UserCollection(app.session()).add(
            username='foo@bar.org',
            password='foobar',
            role='member',
            active=False,
            authentication_provider={
                'name': 'kerberos',
                'data': {
                    'username': 'foo@EXAMPLE.ORG'
                },
                'required': True
            }
        )
        transaction.commit()

        r = auth({'Authorization': 'Negotiate foobar'})
        assert r.status_code == 403

        # now activate the user
        user = UserCollection(app.session()).by_username('foo@bar.org')
        user.active = True
        transaction.commit()

        r = auth({'Authorization': 'Negotiate foobar'})
        assert r.status_code == 302
        assert r.location == 'http://localhost/'
        assert r.headers['Set-Cookie'].startswith('session_id')

        # since the user *requires* the auth provider, a normal login
        # is no longer possible
        auth = Auth.from_app(app)
        assert not auth.authenticate('foo@bar.org', password='foobar')

        # unless we make it optional
        user = UserCollection(app.session()).by_username('foo@bar.org')
        user.authentication_provider['required'] = False
        transaction.commit()

        auth = Auth.from_app(app)
        assert auth.authenticate('foo@bar.org', password='foobar')



from onegov.user.utils import password_reset_url
from onegov.user.collections import UserCollection


class DummyRequest():
    client_addr = '127.0.0.1'

    def new_url_safe_token(self, token):
        return '-'.join(['_'.join(pair) for pair in sorted(token.items())])


def test_password_reset_url(session):
    request = DummyRequest()
    user = UserCollection(session).register('usr', 'very_secret', request)

    url = password_reset_url(user, request, 'http://localhost/reset')
    assert url == 'http://localhost/reset?token=modified_-username_usr'

    url = password_reset_url(user, request, 'http://localhost/reset?extra')
    assert url == 'http://localhost/reset?extra&token=modified_-username_usr'

    url = password_reset_url(user, request, 'http://localhost/reset?p=1&p=2')
    assert url == 'http://localhost/reset?p=1&p=2&token=modified_-username_usr'



import kerberos
import os

from abc import ABCMeta, abstractmethod
from attr import attrs, attrib
from attr.validators import in_
from contextlib import contextmanager
from onegov.form import WTFormsClassBuilder
from onegov.user import _, log
from onegov.user.models.user import User
from webob.exc import HTTPUnauthorized
from wtforms.fields import BooleanField, RadioField, StringField
from translationstring import TranslationString
from typing import Optional
from ua_parser import user_agent_parser


AUTHENTICATION_PROVIDERS = {}


def provider_by_name(providers, name):
    return next((p for p in providers if p.metadata.name == name), None)


@attrs(slots=True, frozen=True)
class ProviderMetadata(object):
    """ Holds provider-specific metadata. """

    name: str = attrib()
    title: str = attrib()


@attrs(slots=True, frozen=True)
class UserField(object):
    """ Defines user-specific field.

    The following properties are required:

        * suffix => part of the attribute name ([a-z_]+)
        * label => the translatable label of the field
        * type => the type of the field (currently only 'string')

    """

    field_classes = {'string': StringField}

    suffix: str = attrib()
    label: TranslationString = attrib()
    type: str = attrib(validator=in_(field_classes.keys()))

    @property
    def field_class(self):
        return self.field_classes[self.type]


def include_provider_form_fields(providers, form_class):
    """ Extends a form_class with provider selection.

    The form class contains a list of providers to chose from and the
    configuration necessary on the user for each provider. The providers
    list is expected to be a list of provider instances.

    The resulting data is automatically applied if the user is used
    as a base for the model (via the authentication_provider property).

    The form class is always returned with a new authentication_provider
    that can be stored on the user property of the same name.

    This property is available even if there are no providers - in this case
    it will always return None.

    """

    class AuthProviderForm(form_class):

        @property
        def authentication_provider(self):
            provider = provider_by_name(providers, self.provider.data)

            if not provider:
                return None

            fields = {}

            for field in provider.user_fields:
                form_field = f'{provider.metadata.name}_{field.suffix}'
                fields[field.suffix] = getattr(self, form_field).data

            return {
                'name': provider.metadata.name,
                'fields': fields,
                'required': self.provider_required.data,
            }

        @authentication_provider.setter
        def authentication_provider(self, data):
            self.provider.data = data['name']
            self.provider_required.data = data['required']

            for key, value in data['fields'].items():
                form_field = f"{data['name']}_{key}"
                getattr(self, form_field).data = value

        def populate_obj(self, model):
            super().populate_obj(model)

            if providers:
                model.authentication_provider = self.authentication_provider

        def process_obj(self, model):
            super().process_obj(model)

            if providers and model.authentication_provider:
                self.authentication_provider = model.authentication_provider

    if not providers:
        return AuthProviderForm

    choices = [('none', _("None"))]
    choices.extend((p.metadata.name, p.metadata.title) for p in providers)

    builder = WTFormsClassBuilder(base_class=AuthProviderForm)
    builder.set_current_fieldset(_("Third-Party Authentication"))

    builder.add_field(
        field_class=RadioField,
        field_id='provider',
        label=_("Provider"),
        required=False,
        default='none',
        choices=choices
    )

    for provider in providers:
        for field in provider.user_fields:

            builder.add_field(
                field_class=field.field_class,
                field_id=f'{provider.metadata.name}_{field.suffix}',
                label=field.label,
                required=True,
                depends_on=('provider', provider.metadata.name)
            )

    builder.add_field(
        field_class=BooleanField,
        field_id='provider_required',
        label=_("Force login through provider"),
        required=False,
        description=_(
            "Forces the user to use this provider. Regular username/password "
            "authentication will be disabled!"
        ),
        depends_on=('provider', '!none'))

    return builder.form_class


@attrs()
class AuthenticationProvider(metaclass=ABCMeta):
    """ Base class and registry for third party authentication providers. """

    # stores the 'to' attribute for the integration app
    # :class:`~onegov.user.integration.UserApp`.
    to: Optional[str] = attrib(init=False)

    @property
    def name(self):
        """ Needs to be available for the path in the integration app. """
        return self.metadata.name

    def __init_subclass__(cls, metadata, **kwargs):
        global AUTHENTICATION_PROVIDERS
        assert metadata.name not in AUTHENTICATION_PROVIDERS

        cls.metadata = metadata
        AUTHENTICATION_PROVIDERS[metadata.name] = cls

        super().__init_subclass__(**kwargs)

    @abstractmethod
    def authenticate_request(self, request):
        """ Authenticates the given request in one or many steps.

        Providers are expected to return one of the following values:

        * A valid user (if the authentication was successful)
        * None (if the authentication failed)
        * A webob response (to perform handshakes)

        This function is called whenever the authentication is initiated by
        the user. If the provider returns a webob response, it is returned
        as is by the calling view.

        Therefore, `authenticate_request` must take care to return responses
        in a way that eventually end up fulfilling the authentication. At the
        very least, providers should ensure that all parameters of the original
        request are kept when asking external services to call back.

        """

    @abstractmethod
    def button_text(self, request):
        """ Returns the translatable button text for the given request.

        It is okay to return a static text, if the button remains the same
        for all requests.

        The translatable text is parsed as markdown, to add weight to
        the central element of the text. For example::

            Login with **Windows**

        """

    def available_users(self, request):
        """ Returns a query limited to users which may be authenticated
        using the given provider.

        This should be used as a base for the identification, rather than
        building your own query, as inactive users and ones without the
        proper configuration are excluded.

        """
        return request.session.query(User)\
            .filter_by(active=True)\
            .filter(User.authentication_provider['name'] == self.metadata.name)

    @classmethod
    def configure(cls, **kwargs):
        """ This function gets called with the per-provider configuration
        defined in onegov.yml. Authentication providers may optionally
        access these values.

        The return value is either a provider instance, or none if the
        provider is not available.

        """

        return cls()

    @property
    def user_fields(self, request):
        """ Optional fields required by the provider on the user. Should return
        something that is iterable (even if only one or no fields are used).

        See :class:`UserField`.
        """

        return ()


@attrs()
class KerberosProvider(AuthenticationProvider, metadata=ProviderMetadata(
    name='kerberos', title=_("Kerberos (v5)")
)):
    """ Kerberos is a computer-network authentication protocol that works on
    the basis of tickets to allow nodes communicating over a non-secure network
    to prove their identity to one another in a secure manner.

    """

    keytab: str = attrib()
    hostname: str = attrib()
    service: str = attrib()

    @property
    def user_fields(self):
        yield UserField('username', _("Kerberos username"), 'string')

    @classmethod
    def configure(cls, **cfg):
        keytab = cfg.get('keytab', None)
        hostname = cfg.get('hostname', None)
        service = cfg.get('service', None)

        if not keytab:
            return None

        if not hostname:
            return None

        if not service:
            return None

        provider = cls(keytab, hostname, service)

        with provider.context():
            try:
                kerberos.getServerPrincipalDetails(
                    provider.service, provider.hostname)
            except kerberos.KrbError as e:
                log.warning(f"Kerberos config error: {e}")
            else:
                return provider

    @contextmanager
    def context(self):
        """ Runs the block inside the context manager with the keytab
        set to the provider's keytab.

        All functions that interact with kerberos must be run inside
        this context.

        For convenience, this context returns the kerberos module
        when invoked.

        """
        previous = os.environ.pop('KRB5_KTNAME', None)
        os.environ['KRB5_KTNAME'] = self.keytab

        yield

        if previous is not None:
            os.environ['KRB5_KTNAME'] = previous

    def button_text(self, request):
        """ Returns the request tailored to each OS (users won't understand
        Kerberos, but for them it's basically their local OS login).

        """
        agent = user_agent_parser.Parse(request.user_agent or "")
        agent_os = agent['os']['family']

        if agent_os == "Other":
            return _("Login with operating system")

        return _("Login with **${operating_system}**", mapping={
            'operating_system': agent_os
        })

    def authenticate_request(self, request):
        """ Authenticates the kerberos request.

        The kerberos handshake is as follows:

        1. An HTTPUnauthorized response (401) is returned, with the
           WWW-Authenticate header set to "Negotiate"

        2. The client sends a request with the Authorization header set
           to the kerberos ticket.

        """

        # extract the token
        token = request.headers.get('Authorization')
        token = token and ''.join(token.split()[1:]).strip()

        def with_header(response, include_token=True):
            if include_token and token:
                negotiate = f'Negotiate {token}'
            else:
                negotiate = 'Negotiate'

            response.headers['WWW-Authenticate'] = negotiate

            return response

        def negotiate():
            # only mirror the token back, if it is valid, which is never
            # the case in the negotiate step
            return with_header(HTTPUnauthorized(), include_token=False)

        # ask for a token
        if not token:
            return negotiate()

        # verify the token
        with self.context():

            # initialization step
            result, state = kerberos.authGSSServerInit(self.service)

            if result != kerberos.AUTH_GSS_COMPLETE:
                return negotiate()

            # challenge step
            result = kerberos.authGSSServerStep(state, token)

            if result != kerberos.AUTH_GSS_COMPLETE:
                return negotiate()

            # extract the final token
            token = kerberos.authGSSServerResponse(state)

            # include the token in the response
            request.after(with_header)

            # extract the user if possible
            username = kerberos.authGSSServerUserName(state)
            selector = User.authentication_provider['data']['username']

            return self.available_users(request)\
                .filter(selector == username)\
                .first()



from onegov.core.orm import Base
from onegov.core.orm.mixins import TimestampMixin
from onegov.core.orm.mixins import ContentMixin
from onegov.core.orm.types import UUID
from sqlalchemy import Column
from sqlalchemy import Text
from uuid import uuid4


class UserGroup(Base, ContentMixin, TimestampMixin):
    """ Defines a generic user group. """

    __tablename__ = 'groups'

    #: the type of the item, this can be used to create custom polymorphic
    #: subclasses of this class. See
    #: `<http://docs.sqlalchemy.org/en/improve_toc/\
    #: orm/extensions/declarative/inheritance.html>`_.
    type = Column(Text, nullable=True)

    __mapper_args__ = {
        'polymorphic_on': type
    }

    #: the id of the user group
    id = Column(UUID, primary_key=True, default=uuid4)

    #: the name of this group
    name = Column(Text, nullable=True)



from abc import ABCMeta, abstractmethod
from onegov.core.utils import is_valid_yubikey


SECOND_FACTORS = {}


class SecondFactor(metaclass=ABCMeta):
    """ Base class and registry for secondary auth factors. """

    def __init_subclass__(cls, type, **kwargs):
        global SECOND_FACTORS
        assert type not in SECOND_FACTORS

        SECOND_FACTORS[type] = cls

        super().__init_subclass__(**kwargs)

    @abstractmethod
    def __init__(self, **kwargs):
        """ Initialises the auth factor using a dictionary that may or may
        not contain the configuration values necessary for the auth factor.

        Either way, the auth factor is instantiated even if the needed
        configuration is not available.

        The :meth:`is_configured` method is responsible for signaling if
        the configuration has been successful or not.

        All used configuration values should be popped, not just read.

        """

    @classmethod
    @abstractmethod
    def args_from_app(cls, app):
        """ Copies the required configuration values from the app, returning
        a dictionary with all keys present. The values should be either the
        ones from the application or None.

        """

    @abstractmethod
    def is_configured(self):
        """ Returns `True` if the factor has been properly configured. """

    @abstractmethod
    def is_valid(self, user_specific_config, factor):
        """ Returns true if the given factor is valid for the given
        user-specific configuration. This is the value stored on the
        user in the `second_factor` column.

        """


class YubikeyFactor(SecondFactor, type='yubikey'):
    """ Implements a yubikey factor for the :class:`Auth` class. """

    def __init__(self, **kwargs):
        self.yubikey_client_id = kwargs.pop('yubikey_client_id', None)
        self.yubikey_secret_key = kwargs.pop('yubikey_secret_key', None)

    @classmethod
    def args_from_app(cls, app):
        return {
            'yubikey_client_id': getattr(app, 'yubikey_client_id', None),
            'yubikey_secret_key': getattr(app, 'yubikey_secret_key', None)
        }

    def is_configured(self):
        return self.yubikey_client_id and self.yubikey_secret_key

    def is_valid(self, user_specific_config, factor):
        return is_valid_yubikey(
            client_id=self.yubikey_client_id,
            secret_key=self.yubikey_secret_key,
            expected_yubikey_id=user_specific_config,
            yubikey=factor
        )



from morepath import App
from onegov.core.cache import lru_cache
from onegov.core.security import Public
from onegov.user.auth.core import Auth
from onegov.user.auth.provider import AUTHENTICATION_PROVIDERS
from onegov.user.auth.provider import AuthenticationProvider
from onegov.user.auth.provider import provider_by_name
from onegov.user.models.user import User
from webob.exc import HTTPForbidden
from webob.response import Response


class UserApp(App):
    """ Provides user integration.

    Historically it was not necessary to use this app for user integration,
    and most features are still possible without it. However, third-party
    authentication providers only work if the UserApp is integrated.

    The following configuration options are accepted:

    :authentication_providers:

        A dictionary of provider-specific configuration settings, see
        :mod:`onegov.user.auth.provider` for more information.

    """

    @property
    def providers(self):
        """ Returns a tuple of availabe providers. """

        return getattr(self, 'available_providers', ())

    @lru_cache(maxsize=8)
    def provider(self, name):
        return provider_by_name(self.providers, name)

    def configure_authentication_providers(self, **cfg):

        def bound(provider):
            if 'authentication_providers' not in cfg:
                return {}

            if provider.metadata.name not in cfg['authentication_providers']:
                return {}

            return cfg['authentication_providers'][provider.metadata.name]

        available = AUTHENTICATION_PROVIDERS.values()
        available = (cls.configure(**bound(cls)) for cls in available)
        available = (obj for obj in available if obj is not None)

        self.available_providers = tuple(available)


@UserApp.path(
    model=AuthenticationProvider,
    path='/auth/provider/{name}')
def authentication_provider(app, name, to='/'):
    provider = app.provider(name)

    if not provider:
        return None

    # the 'to' is just held here to be able to reuse it in the view
    provider.to = to

    return provider


@UserApp.view(
    model=AuthenticationProvider,
    permission=Public)
def handle_authentication(self, request):
    response = self.authenticate_request(request)

    # the provider failed to authenticate
    if response is None:
        return HTTPForbidden()

    # the provider returned its own response
    if isinstance(response, Response):
        return response

    # the provider authenticated a user
    if isinstance(response, User):
        return Auth.from_request(request, to=self.to)\
            .complete_login(user=response, request=request)

    # the provider returned something illegal
    raise RuntimeError(
        f"Invalid response from {self.metadata.name}: {response}")
